package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching;
import org.eclipse.jdt.internal.compiler.ast.ReferenceExpression;
import org.eclipse.jdt.internal.compiler.ast.Invocation;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;

public class ParameterizedGenericMethodBinding extends ParameterizedMethodBinding implements Substitution
{
    public TypeBinding[] typeArguments;
    protected LookupEnvironment environment;
    public boolean inferredReturnType;
    public boolean wasInferred;
    public boolean isRaw;
    private MethodBinding tiebreakMethod;
    public boolean inferredWithUncheckedConversion;
    private static /* synthetic */ int[] $SWITCH_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus;
    
    public static MethodBinding computeCompatibleMethod(MethodBinding originalMethod, TypeBinding[] arguments, Scope scope, InvocationSite invocationSite) {
        TypeVariableBinding[] typeVariables = originalMethod.typeVariables;
        TypeBinding[] substitutes = invocationSite.genericTypeArguments();
        InferenceContext inferenceContext = null;
        TypeBinding[] uncheckedArguments = null;
        ParameterizedGenericMethodBinding methodSubstitute;
        if (substitutes != null) {
            if (substitutes.length != typeVariables.length) {
                return new ProblemMethodBinding(originalMethod, originalMethod.selector, substitutes, 11);
            }
            methodSubstitute = scope.environment().createParameterizedGenericMethod(originalMethod, substitutes);
        }
        else {
            TypeBinding[] parameters = originalMethod.parameters;
            CompilerOptions compilerOptions = scope.compilerOptions();
            if (compilerOptions.sourceLevel >= 3407872L) {
                return computeCompatibleMethod18(originalMethod, arguments, scope, invocationSite);
            }
            inferenceContext = new InferenceContext(originalMethod);
            methodSubstitute = inferFromArgumentTypes(scope, originalMethod, arguments, parameters, inferenceContext);
            if (methodSubstitute == null) {
                return null;
            }
            if (inferenceContext.hasUnresolvedTypeArgument()) {
                if (inferenceContext.isUnchecked) {
                    int length = inferenceContext.substitutes.length;
                    System.arraycopy(inferenceContext.substitutes, 0, uncheckedArguments = new TypeBinding[length], 0, length);
                }
                if (methodSubstitute.returnType != TypeBinding.VOID) {
                    TypeBinding expectedType = invocationSite.invocationTargetType();
                    if (expectedType != null) {
                        inferenceContext.hasExplicitExpectedType = true;
                    }
                    else {
                        expectedType = scope.getJavaLangObject();
                    }
                    inferenceContext.expectedType = expectedType;
                }
                methodSubstitute = methodSubstitute.inferFromExpectedType(scope, inferenceContext);
                if (methodSubstitute == null) {
                    return null;
                }
            }
            else if (compilerOptions.sourceLevel == 3342336L && methodSubstitute.returnType != TypeBinding.VOID) {
                TypeBinding expectedType = invocationSite.invocationTargetType();
                if (expectedType != null && !originalMethod.returnType.mentionsAny(originalMethod.parameters, -1)) {
                    TypeBinding uncaptured = methodSubstitute.returnType.uncapture(scope);
                    if (!methodSubstitute.returnType.isCompatibleWith(expectedType) && expectedType.isCompatibleWith(uncaptured)) {
                        InferenceContext oldContext = inferenceContext;
                        inferenceContext = new InferenceContext(originalMethod);
                        originalMethod.returnType.collectSubstitutes(scope, expectedType, inferenceContext, 1);
                        ParameterizedGenericMethodBinding substitute = inferFromArgumentTypes(scope, originalMethod, arguments, parameters, inferenceContext);
                        if (substitute != null && substitute.returnType.isCompatibleWith(expectedType)) {
                            if (scope.parameterCompatibilityLevel(substitute, arguments, false) > -1) {
                                methodSubstitute = substitute;
                            }
                            else {
                                inferenceContext = oldContext;
                            }
                        }
                        else {
                            inferenceContext = oldContext;
                        }
                    }
                }
            }
        }
        Substitution substitution = null;
        if (inferenceContext != null) {
            substitution = new ParameterizedGenericMethodBinding.LingeringTypeVariableEliminator(typeVariables, inferenceContext.substitutes, scope);
        }
        else {
            substitution = methodSubstitute;
        }
        int i;
        for (i = 0, int length = typeVariables.length; i < length; ++i) {
            TypeVariableBinding typeVariable = typeVariables[i];
            TypeBinding substitute2 = methodSubstitute.typeArguments[i];
            TypeBinding substituteForChecks;
            if ((substitute2 instanceof TypeVariableBinding)) {
                substituteForChecks = substitute2;
            }
            else {
                substituteForChecks = Scope.substitute(new ParameterizedGenericMethodBinding.LingeringTypeVariableEliminator(typeVariables, null, scope), substitute2);
            }
            if (uncheckedArguments == null || uncheckedArguments[i] != null) {
                switch ($SWITCH_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus()[typeVariable.boundCheck(substitution, substituteForChecks, scope, null).ordinal()]) {
                    case 4: {
                        int argLength = arguments.length;
                        TypeBinding[] augmentedArguments = new TypeBinding[argLength + 2];
                        System.arraycopy(arguments, 0, augmentedArguments, 0, argLength);
                        augmentedArguments[argLength] = substitute2;
                        augmentedArguments[argLength + 1] = typeVariable;
                        return new ProblemMethodBinding(methodSubstitute, originalMethod.selector, augmentedArguments, 10);
                    }
                    case 3: {
                        methodSubstitute.tagBits |= 0x100L;
                        break;
                    }
                }
            }
        }
        return methodSubstitute;
    }
    
    public static MethodBinding computeCompatibleMethod18(MethodBinding originalMethod, TypeBinding[] arguments, Scope scope, InvocationSite invocationSite) {
        TypeVariableBinding[] typeVariables = originalMethod.typeVariables;
        if (invocationSite.checkingPotentialCompatibility()) {
            return scope.environment().createParameterizedGenericMethod(originalMethod, typeVariables);
        }
        ParameterizedGenericMethodBinding methodSubstitute = null;
        InferenceContext18 infCtx18 = invocationSite.freshInferenceContext(scope);
        if (infCtx18 == null) {
            return originalMethod;
        }
        TypeBinding[] parameters = originalMethod.parameters;
        CompilerOptions compilerOptions = scope.compilerOptions();
        boolean invocationTypeInferred = false;
        boolean requireBoxing = false;
        TypeBinding[] argumentsCopy = new TypeBinding[arguments.length];
        int i = 0;
        int length = arguments.length;
        int parametersLength = parameters.length;
        while (i < length) {
            TypeBinding parameter = (i < parametersLength) ? parameters[i] : parameters[parametersLength - 1];
            TypeBinding argument = arguments[i];
            if (argument.isPrimitiveType() != parameter.isPrimitiveType()) {
                argumentsCopy[i] = scope.environment().computeBoxingType(argument);
                requireBoxing = true;
            }
            else {
                argumentsCopy[i] = argument;
            }
            ++i;
        }
        arguments = argumentsCopy;
        LookupEnvironment environment = scope.environment();
        InferenceContext18 previousContext = environment.currentInferenceContext;
        if (previousContext == null) {
            environment.currentInferenceContext = infCtx18;
        }
        try {
            BoundSet provisionalResult = null;
            BoundSet result = null;
            boolean isPolyExpression = invocationSite instanceof Expression && ((Expression)invocationSite).isPolyExpression(originalMethod);
            boolean isDiamond = isPolyExpression && originalMethod.isConstructor();
            if (arguments.length == parameters.length) {
                infCtx18.inferenceKind = (requireBoxing ? 2 : 1);
                infCtx18.inferInvocationApplicability(originalMethod, arguments, isDiamond);
                result = infCtx18.solve(true);
            }
            if (result == null && originalMethod.isVarargs()) {
                infCtx18 = invocationSite.freshInferenceContext(scope);
                infCtx18.inferenceKind = 3;
                infCtx18.inferInvocationApplicability(originalMethod, arguments, isDiamond);
                result = infCtx18.solve(true);
            }
            if (result == null) {
                return null;
            }
            if (infCtx18.isResolved(result)) {
                infCtx18.stepCompleted = 1;
                TypeBinding expectedType = invocationSite.invocationTargetType();
                boolean hasReturnProblem = false;
                if (expectedType != null || !invocationSite.getExpressionContext().definesTargetType() || !isPolyExpression) {
                    provisionalResult = result;
                    result = infCtx18.inferInvocationType(expectedType, invocationSite, originalMethod);
                    invocationTypeInferred = true;
                    hasReturnProblem |= (result == null);
                    if (hasReturnProblem) {
                        result = provisionalResult;
                    }
                }
                if (result != null) {
                    TypeBinding[] solutions = infCtx18.getSolutions(typeVariables, invocationSite, result);
                    if (solutions != null) {
                        methodSubstitute = scope.environment().createParameterizedGenericMethod(originalMethod, solutions, infCtx18.usesUncheckedConversion, hasReturnProblem);
                        if ((invocationSite instanceof Invocation)) {
                            infCtx18.forwardResults(result, (Invocation)invocationSite, methodSubstitute, expectedType);
                        }
                        try {
                            if (hasReturnProblem) {
                                MethodBinding problemMethod = infCtx18.getReturnProblemMethodIfNeeded(expectedType, methodSubstitute);
                                if ((problemMethod instanceof ProblemMethodBinding)) {
                                    if ((invocationSite instanceof Invocation)) {
                                        ((Invocation)invocationSite).registerInferenceContext(methodSubstitute, infCtx18);
                                    }
                                    else if ((invocationSite instanceof ReferenceExpression)) {
                                        ((ReferenceExpression)invocationSite).registerInferenceContext(methodSubstitute, infCtx18);
                                    }
                                    return problemMethod;
                                }
                            }
                            if (invocationTypeInferred) {
                                if (compilerOptions.isAnnotationBasedNullAnalysisEnabled) {
                                    NullAnnotationMatching.checkForContradictions(methodSubstitute, invocationSite, scope);
                                }
                                MethodBinding problemMethod = methodSubstitute.boundCheck18(scope, arguments, invocationSite);
                                if (problemMethod != null) {
                                    if ((invocationSite instanceof Invocation)) {
                                        ((Invocation)invocationSite).registerInferenceContext(methodSubstitute, infCtx18);
                                    }
                                    else if ((invocationSite instanceof ReferenceExpression)) {
                                        ((ReferenceExpression)invocationSite).registerInferenceContext(methodSubstitute, infCtx18);
                                    }
                                    return problemMethod;
                                }
                            }
                            else {
                                methodSubstitute = new PolyParameterizedGenericMethodBinding(methodSubstitute);
                            }
                        }
                        finally {
                            if ((invocationSite instanceof Invocation)) {
                                ((Invocation)invocationSite).registerInferenceContext(methodSubstitute, infCtx18);
                            }
                            else if ((invocationSite instanceof ReferenceExpression)) {
                                ((ReferenceExpression)invocationSite).registerInferenceContext(methodSubstitute, infCtx18);
                            }
                        }
                        return methodSubstitute;
                    }
                }
                return null;
            }
            else {
                return null;
            }
        }
        catch (InferenceFailureException e) {
            scope.problemReporter().genericInferenceError(e.getMessage(), invocationSite);
            return null;
        }
        finally {
            environment.currentInferenceContext = previousContext;
        }
    }
    
    MethodBinding boundCheck18(Scope scope, TypeBinding[] arguments, InvocationSite site) {
        Substitution substitution = this;
        ParameterizedGenericMethodBinding methodSubstitute = this;
        TypeVariableBinding[] originalTypeVariables = this.originalMethod.typeVariables;
        for (int i = 0, length = originalTypeVariables.length; i < length; ++i) {
            TypeVariableBinding typeVariable = originalTypeVariables[i];
            TypeBinding substitute = methodSubstitute.typeArguments[i];
            ASTNode location = (site instanceof ASTNode) ? ((ASTNode)site) : null;
            switch ($SWITCH_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus()[typeVariable.boundCheck(substitution, substitute, scope, location).ordinal()]) {
                case 4: {
                    int argLength = arguments.length;
                    TypeBinding[] augmentedArguments = new TypeBinding[argLength + 2];
                    System.arraycopy(arguments, 0, augmentedArguments, 0, argLength);
                    augmentedArguments[argLength] = substitute;
                    augmentedArguments[argLength + 1] = typeVariable;
                    return new ProblemMethodBinding(methodSubstitute, this.originalMethod.selector, augmentedArguments, 10);
                }
                case 3: {
                    methodSubstitute.tagBits |= 0x100L;
                    break;
                }
            }
        }
        return null;
    }
    
    private static ParameterizedGenericMethodBinding inferFromArgumentTypes(Scope scope, MethodBinding originalMethod, TypeBinding[] arguments, TypeBinding[] parameters, InferenceContext inferenceContext) {
        if (originalMethod.isVarargs()) {
            int paramLength = parameters.length;
            int minArgLength = paramLength - 1;
            int argLength = arguments.length;
            int j = 0;
            while (j < minArgLength) {
                parameters[j].collectSubstitutes(scope, arguments[j], inferenceContext, 1);
                if (inferenceContext.status == 1) {
                    return null;
                }
                ++j;
            }
            if (minArgLength < argLength) {
                TypeBinding varargType = parameters[minArgLength];
                TypeBinding lastArgument = arguments[minArgLength];
                Label_0158: {
                    if (paramLength == argLength) {
                        if (lastArgument == TypeBinding.NULL) {
                            break Label_0158;
                        }
                        switch (lastArgument.dimensions()) {
                            case 0: {
                                break;
                            }
                            case 1: {
                                if (lastArgument.leafComponentType().isBaseType()) {
                                    break;
                                }
                            }
                            default: {
                                break Label_0158;
                            }
                        }
                    }
                    varargType = ((ArrayBinding)varargType).elementsType();
                }
                int k = minArgLength;
                while (k < argLength) {
                    varargType.collectSubstitutes(scope, arguments[k], inferenceContext, 1);
                    if (inferenceContext.status == 1) {
                        return null;
                    }
                    ++k;
                }
            }
        }
        else {
            int paramLength = parameters.length;
            int i = 0;
            while (i < paramLength) {
                parameters[i].collectSubstitutes(scope, arguments[i], inferenceContext, 1);
                if (inferenceContext.status == 1) {
                    return null;
                }
                ++i;
            }
        }
        TypeVariableBinding[] originalVariables = originalMethod.typeVariables;
        if (!resolveSubstituteConstraints(scope, originalVariables, inferenceContext, false)) {
            return null;
        }
        TypeBinding[] actualSubstitutes;
        TypeBinding[] inferredSustitutes = actualSubstitutes = inferenceContext.substitutes;
        for (int j = 0, varLength = originalVariables.length; j < varLength; ++j) {
            if (inferredSustitutes[j] == null) {
                if (actualSubstitutes == inferredSustitutes) {
                    System.arraycopy(inferredSustitutes, 0, actualSubstitutes = new TypeBinding[varLength], 0, j);
                }
                actualSubstitutes[j] = originalVariables[j];
            }
            else if (actualSubstitutes != inferredSustitutes) {
                actualSubstitutes[j] = inferredSustitutes[j];
            }
        }
        ParameterizedGenericMethodBinding paramMethod = scope.environment().createParameterizedGenericMethod(originalMethod, actualSubstitutes);
        return paramMethod;
    }
    
    private static boolean resolveSubstituteConstraints(Scope scope, TypeVariableBinding[] typeVariables, InferenceContext inferenceContext, boolean considerEXTENDSConstraints) {
        TypeBinding[] substitutes = inferenceContext.substitutes;
        int varLength = typeVariables.length;
    Label_0168:
        for (int i = 0; i < varLength; ++i) {
            TypeVariableBinding current = typeVariables[i];
            TypeBinding substitute = substitutes[i];
            if (substitute == null) {
                TypeBinding[] equalSubstitutes = inferenceContext.getSubstitutes(current, 0);
                if (equalSubstitutes != null) {
                    int j = 0;
                    int equalLength = equalSubstitutes.length;
                    while (j < equalLength) {
                        TypeBinding equalSubstitute = equalSubstitutes[j];
                        if (equalSubstitute == null) {
                            ++j;
                        }
                        else {
                            if (TypeBinding.equalsEquals(equalSubstitute, current)) {
                                int k = j + 1;
                                while (k < equalLength) {
                                    equalSubstitute = equalSubstitutes[k];
                                    if (TypeBinding.notEquals(equalSubstitute, current) && equalSubstitute != null) {
                                        substitutes[i] = equalSubstitute;
                                        continue Label_0168;
                                    }
                                    ++k;
                                }
                                substitutes[i] = current;
                                break;
                            }
                            substitutes[i] = equalSubstitute;
                            break;
                        }
                    }
                }
            }
        }
        if (inferenceContext.hasUnresolvedTypeArgument()) {
            for (int i = 0; i < varLength; ++i) {
                TypeVariableBinding current = typeVariables[i];
                TypeBinding substitute = substitutes[i];
                if (substitute == null) {
                    TypeBinding[] bounds = inferenceContext.getSubstitutes(current, 2);
                    if (bounds != null) {
                        TypeBinding mostSpecificSubstitute = scope.lowerUpperBound(bounds);
                        if (mostSpecificSubstitute == null) {
                            return false;
                        }
                        if (mostSpecificSubstitute != TypeBinding.VOID) {
                            substitutes[i] = mostSpecificSubstitute;
                        }
                    }
                }
            }
        }
        if (considerEXTENDSConstraints && inferenceContext.hasUnresolvedTypeArgument()) {
            for (int i = 0; i < varLength; ++i) {
                TypeVariableBinding current = typeVariables[i];
                TypeBinding substitute = substitutes[i];
                if (substitute == null) {
                    TypeBinding[] bounds = inferenceContext.getSubstitutes(current, 1);
                    if (bounds != null) {
                        TypeBinding[] glb = Scope.greaterLowerBound(bounds, scope, scope.environment());
                        TypeBinding mostSpecificSubstitute2 = null;
                        if (glb != null) {
                            if (glb.length == 1) {
                                mostSpecificSubstitute2 = glb[0];
                            }
                            else {
                                TypeBinding[] otherBounds = new TypeBinding[glb.length - 1];
                                System.arraycopy(glb, 1, otherBounds, 0, glb.length - 1);
                                mostSpecificSubstitute2 = scope.environment().createWildcard(null, 0, glb[0], otherBounds, 1);
                            }
                        }
                        if (mostSpecificSubstitute2 != null) {
                            substitutes[i] = mostSpecificSubstitute2;
                        }
                    }
                }
            }
        }
        return true;
    }
    
    public ParameterizedGenericMethodBinding(MethodBinding originalMethod, RawTypeBinding rawType, LookupEnvironment environment) {
        TypeVariableBinding[] originalVariables = originalMethod.typeVariables;
        int length = originalVariables.length;
        TypeBinding[] rawArguments = new TypeBinding[length];
        for (int i = 0; i < length; ++i) {
            rawArguments[i] = environment.convertToRawType(originalVariables[i].erasure(), false);
        }
        this.isRaw = true;
        this.tagBits = originalMethod.tagBits;
        this.environment = environment;
        this.modifiers = originalMethod.modifiers;
        this.selector = originalMethod.selector;
        this.declaringClass = ((rawType == null) ? originalMethod.declaringClass : rawType);
        this.typeVariables = Binding.NO_TYPE_VARIABLES;
        this.typeArguments = rawArguments;
        this.originalMethod = originalMethod;
        boolean ignoreRawTypeSubstitution = rawType == null || originalMethod.isStatic();
        this.parameters = Scope.substitute(this, ignoreRawTypeSubstitution ? originalMethod.parameters : Scope.substitute(rawType, originalMethod.parameters));
        this.thrownExceptions = Scope.substitute(this, ignoreRawTypeSubstitution ? originalMethod.thrownExceptions : Scope.substitute(rawType, originalMethod.thrownExceptions));
        if (this.thrownExceptions == null) {
            this.thrownExceptions = Binding.NO_EXCEPTIONS;
        }
        this.returnType = Scope.substitute(this, ignoreRawTypeSubstitution ? originalMethod.returnType : Scope.substitute(rawType, originalMethod.returnType));
        this.wasInferred = false;
        this.parameterNonNullness = originalMethod.parameterNonNullness;
        this.defaultNullness = originalMethod.defaultNullness;
    }
    
    public ParameterizedGenericMethodBinding(MethodBinding originalMethod, TypeBinding[] typeArguments, LookupEnvironment environment, boolean inferredWithUncheckConversion, boolean hasReturnProblem) {
        this.environment = environment;
        this.inferredWithUncheckedConversion = inferredWithUncheckConversion;
        this.modifiers = originalMethod.modifiers;
        this.selector = originalMethod.selector;
        this.declaringClass = originalMethod.declaringClass;
        if (inferredWithUncheckConversion && originalMethod.isConstructor() && this.declaringClass.isParameterizedType()) {
            this.declaringClass = (ReferenceBinding)environment.convertToRawType(this.declaringClass.erasure(), false);
        }
        this.typeVariables = Binding.NO_TYPE_VARIABLES;
        this.typeArguments = typeArguments;
        this.isRaw = false;
        this.tagBits = originalMethod.tagBits;
        this.originalMethod = originalMethod;
        this.parameters = Scope.substitute(this, originalMethod.parameters);
        if (inferredWithUncheckConversion) {
            this.returnType = this.getErasure18_5_2(originalMethod.returnType, environment, hasReturnProblem);
            this.thrownExceptions = new ReferenceBinding[originalMethod.thrownExceptions.length];
            for (int i = 0; i < originalMethod.thrownExceptions.length; ++i) {
                this.thrownExceptions[i] = (ReferenceBinding)this.getErasure18_5_2(originalMethod.thrownExceptions[i], environment, false);
            }
        }
        else {
            this.returnType = Scope.substitute(this, originalMethod.returnType);
            this.thrownExceptions = Scope.substitute(this, originalMethod.thrownExceptions);
        }
        if (this.thrownExceptions == null) {
            this.thrownExceptions = Binding.NO_EXCEPTIONS;
        }
        Label_0398: {
            if ((this.tagBits & 0x80L) == 0x0L) {
                if ((this.returnType.tagBits & 0x80L) != 0x0L) {
                    this.tagBits |= 0x80L;
                }
                else {
                    int i = 0;
                    int max = this.parameters.length;
                    while (i < max) {
                        if ((this.parameters[i].tagBits & 0x80L) != 0x0L) {
                            this.tagBits |= 0x80L;
                            break Label_0398;
                        }
                        ++i;
                    }
                    i = 0;
                    max = this.thrownExceptions.length;
                    while (i < max) {
                        if ((this.thrownExceptions[i].tagBits & 0x80L) != 0x0L) {
                            this.tagBits |= 0x80L;
                            break;
                        }
                        ++i;
                    }
                }
            }
        }
        this.wasInferred = true;
        this.parameterNonNullness = originalMethod.parameterNonNullness;
        this.defaultNullness = originalMethod.defaultNullness;
        for (int len = this.parameters.length, j = 0; j < len; ++j) {
            if (this.parameters[j] == TypeBinding.NULL) {
                long nullBits = originalMethod.parameters[j].tagBits & 0x180000000000000L;
                if (nullBits == 72057594037927936L) {
                    if (this.parameterNonNullness == null) {
                        this.parameterNonNullness = new Boolean[len];
                    }
                    this.parameterNonNullness[j] = Boolean.TRUE;
                }
            }
        }
    }
    
    TypeBinding getErasure18_5_2(TypeBinding type, LookupEnvironment env, boolean substitute) {
        if (substitute) {
            type = Scope.substitute(this, type);
        }
        return env.convertToRawType(type.erasure(), true);
    }
    
    @Override
    public char[] computeUniqueKey(boolean isLeaf) {
        StringBuffer buffer = new StringBuffer();
        buffer.append(this.originalMethod.computeUniqueKey(false));
        buffer.append('%');
        buffer.append('<');
        if (!this.isRaw) {
            for (int length = this.typeArguments.length, i = 0; i < length; ++i) {
                TypeBinding typeArgument = this.typeArguments[i];
                buffer.append(typeArgument.computeUniqueKey(false));
            }
        }
        buffer.append('>');
        int resultLength = buffer.length();
        char[] result = new char[resultLength];
        buffer.getChars(0, resultLength, result, 0);
        return result;
    }
    
    @Override
    public LookupEnvironment environment() {
        return this.environment;
    }
    
    @Override
    public boolean hasSubstitutedParameters() {
        if (this.wasInferred) {
            return this.originalMethod.hasSubstitutedParameters();
        }
        return super.hasSubstitutedParameters();
    }
    
    @Override
    public boolean hasSubstitutedReturnType() {
        if (this.inferredReturnType) {
            return this.originalMethod.hasSubstitutedReturnType();
        }
        return super.hasSubstitutedReturnType();
    }
    
    private ParameterizedGenericMethodBinding inferFromExpectedType(Scope scope, InferenceContext inferenceContext) {
        TypeVariableBinding[] originalVariables = this.originalMethod.typeVariables;
        int varLength = originalVariables.length;
        if (inferenceContext.expectedType != null) {
            this.returnType.collectSubstitutes(scope, inferenceContext.expectedType, inferenceContext, 2);
            if (inferenceContext.status == 1) {
                return null;
            }
        }
        for (int i = 0; i < varLength; ++i) {
            TypeVariableBinding originalVariable = originalVariables[i];
            TypeBinding argument = this.typeArguments[i];
            boolean argAlreadyInferred = TypeBinding.notEquals(argument, originalVariable);
            if (TypeBinding.equalsEquals(originalVariable.firstBound, originalVariable.superclass)) {
                TypeBinding substitutedBound = Scope.substitute(this, originalVariable.superclass);
                argument.collectSubstitutes(scope, substitutedBound, inferenceContext, 2);
                if (inferenceContext.status == 1) {
                    return null;
                }
                if (argAlreadyInferred) {
                    substitutedBound.collectSubstitutes(scope, argument, inferenceContext, 1);
                    if (inferenceContext.status == 1) {
                        return null;
                    }
                }
            }
            int j = 0;
            int max = originalVariable.superInterfaces.length;
            while (j < max) {
                TypeBinding substitutedBound2 = Scope.substitute(this, originalVariable.superInterfaces[j]);
                argument.collectSubstitutes(scope, substitutedBound2, inferenceContext, 2);
                if (inferenceContext.status == 1) {
                    return null;
                }
                if (argAlreadyInferred) {
                    substitutedBound2.collectSubstitutes(scope, argument, inferenceContext, 1);
                    if (inferenceContext.status == 1) {
                        return null;
                    }
                }
                ++j;
            }
        }
        if (!resolveSubstituteConstraints(scope, originalVariables, inferenceContext, true)) {
            return null;
        }
        for (int i = 0; i < varLength; ++i) {
            TypeBinding substitute = inferenceContext.substitutes[i];
            if (substitute != null) {
                this.typeArguments[i] = substitute;
            }
            else {
                this.typeArguments[i] = (inferenceContext.substitutes[i] = originalVariables[i].upperBound());
            }
        }
        this.typeArguments = Scope.substitute(this, this.typeArguments);
        TypeBinding oldReturnType = this.returnType;
        this.returnType = Scope.substitute(this, this.returnType);
        this.inferredReturnType = (inferenceContext.hasExplicitExpectedType && TypeBinding.notEquals(this.returnType, oldReturnType));
        this.parameters = Scope.substitute(this, this.parameters);
        this.thrownExceptions = Scope.substitute(this, this.thrownExceptions);
        if (this.thrownExceptions == null) {
            this.thrownExceptions = Binding.NO_EXCEPTIONS;
        }
        if ((this.tagBits & 0x80L) == 0x0L) {
            if ((this.returnType.tagBits & 0x80L) != 0x0L) {
                this.tagBits |= 0x80L;
            }
            else {
                int k = 0;
                int max2 = this.parameters.length;
                while (k < max2) {
                    if ((this.parameters[k].tagBits & 0x80L) != 0x0L) {
                        this.tagBits |= 0x80L;
                        return this;
                    }
                    ++k;
                }
                k = 0;
                max2 = this.thrownExceptions.length;
                while (k < max2) {
                    if ((this.thrownExceptions[k].tagBits & 0x80L) != 0x0L) {
                        this.tagBits |= 0x80L;
                        break;
                    }
                    ++k;
                }
            }
        }
        return this;
    }
    
    @Override
    public boolean isParameterizedGeneric() {
        return true;
    }
    
    @Override
    public boolean isRawSubstitution() {
        return this.isRaw;
    }
    
    @Override
    public TypeBinding substitute(TypeVariableBinding originalVariable) {
        TypeVariableBinding[] variables = this.originalMethod.typeVariables;
        int length = variables.length;
        if (originalVariable.rank < length && TypeBinding.equalsEquals(variables[originalVariable.rank], originalVariable)) {
            TypeBinding substitute = this.typeArguments[originalVariable.rank];
            return originalVariable.combineTypeAnnotations(substitute);
        }
        return originalVariable;
    }
    
    @Override
    public MethodBinding tiebreakMethod() {
        if (this.tiebreakMethod == null) {
            this.tiebreakMethod = this.originalMethod.asRawMethod(this.environment);
        }
        return this.tiebreakMethod;
    }
    
    @Override
    public MethodBinding genericMethod() {
        if (this.isRaw) {
            return this;
        }
        return this.originalMethod;
    }
    
    static /* synthetic */ int[] $SWITCH_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus() {
        int[] $switch_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus = ParameterizedGenericMethodBinding.$SWITCH_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus;
        if ($switch_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus != null) {
            return $switch_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus;
        }
        int[] $switch_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus2 = new int[TypeConstants.BoundCheckStatus.values().length];
        try {
            $switch_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus2[TypeConstants.BoundCheckStatus.MISMATCH.ordinal()] = 4;
        }
        catch (NoSuchFieldError noSuchFieldError) {}
        try {
            $switch_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus2[TypeConstants.BoundCheckStatus.NULL_PROBLEM.ordinal()] = 2;
        }
        catch (NoSuchFieldError noSuchFieldError2) {}
        try {
            $switch_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus2[TypeConstants.BoundCheckStatus.OK.ordinal()] = 1;
        }
        catch (NoSuchFieldError noSuchFieldError3) {}
        try {
            $switch_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus2[TypeConstants.BoundCheckStatus.UNCHECKED.ordinal()] = 3;
        }
        catch (NoSuchFieldError noSuchFieldError4) {}
        return ParameterizedGenericMethodBinding.$SWITCH_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus = $switch_TABLE$org$eclipse$jdt$internal$compiler$lookup$TypeConstants$BoundCheckStatus2;
    }
    
    private static class LingeringTypeVariableEliminator implements Substitution
    {
        private final TypeVariableBinding[] variables;
        private final TypeBinding[] substitutes;
        private final Scope scope;
        
        public LingeringTypeVariableEliminator(TypeVariableBinding[] variables, TypeBinding[] substitutes, Scope scope) {
            this.variables = variables;
            this.substitutes = substitutes;
            this.scope = scope;
        }
        
        @Override
        public TypeBinding substitute(TypeVariableBinding typeVariable) {
            if (typeVariable.rank >= this.variables.length || TypeBinding.notEquals(this.variables[typeVariable.rank], typeVariable)) {
                return typeVariable;
            }
            if (this.substitutes != null) {
                return Scope.substitute(new ParameterizedGenericMethodBinding.LingeringTypeVariableEliminator(this.variables, null, this.scope), this.substitutes[typeVariable.rank]);
            }
            ReferenceBinding genericType = (ReferenceBinding)((typeVariable.declaringElement instanceof ReferenceBinding) ? typeVariable.declaringElement : null);
            return this.scope.environment().createWildcard(genericType, typeVariable.rank, null, null, 0, typeVariable.getTypeAnnotations());
        }
        
        @Override
        public LookupEnvironment environment() {
            return this.scope.environment();
        }
        
        @Override
        public boolean isRawSubstitution() {
            return false;
        }
    }
}
